package ru.dbotthepony.kstarbound

import com.google.gson.JsonObject
import org.apache.logging.log4j.LogManager
import org.lwjgl.Version
import org.lwjgl.glfw.GLFW.glfwSetWindowShouldClose
import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.collision.shapes.CircleShape
import ru.dbotthepony.kbox2d.collision.shapes.PolygonShape
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.defs.TileDefinition
import ru.dbotthepony.kstarbound.defs.projectile.ProjectilePhysics
import ru.dbotthepony.kstarbound.defs.world.dungeon.DungeonWorldDef
import ru.dbotthepony.kstarbound.io.*
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE_FF
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.entities.Move
import ru.dbotthepony.kstarbound.world.entities.PlayerEntity
import ru.dbotthepony.kstarbound.world.entities.projectile.Projectile
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import java.io.ByteArrayInputStream
import java.io.DataInputStream
import java.io.File
import java.io.InputStream
import java.util.*
import java.util.zip.Inflater

private val LOGGER = LogManager.getLogger()

fun main() {
	LOGGER.info("Running LWJGL ${Version.getVersion()}")

	if (true) {
		//val pak = StarboundPak(File("J:\\SteamLibrary\\steamapps\\common\\Starbound\\assets\\packed.pak"))
		//val json = JsonParser.parseReader(pak.getReader("/projectiles/traps/lowgravboostergas/lowgravboostergas.projectile"))
		//val obj = Gson().fromJson(json, ProjectileDefinitionBuilder::class.java)
		//println(obj.build())
		//return
	}

	val db = BTreeDB(File("F:\\SteamLibrary\\steamapps\\common\\Starbound - Unstable\\storage\\universe\\389760395_938904237_-238610574_5.world"))

	/*if (true) {
		val a = System.currentTimeMillis()
		val worldMeta = db.read(byteArrayOf(0, 0, 0, 0, 0))
		println(System.currentTimeMillis() - a)

		val inflater = Inflater()
		inflater.setInput(worldMeta!!)

		val output = ByteArray(1_000_000)
		inflater.inflate(output)

		val stream = DataInputStream(ByteArrayInputStream(output))
		println("X tiles ${stream.readInt()}")
		println("Y tiles ${stream.readInt()}")

		val metadata = VersionedJSON(stream)
		println(metadata.data)

		return
	}*/

	/*if (true) {
		val data = db.read(byteArrayOf(1, 0, 61, 0, 23))

		val inflater = Inflater()
		inflater.setInput(data!!)

		val output = ByteArray(64_000)
		val actual = inflater.inflate(output)
		File("F:\\SteamLibrary\\steamapps\\common\\Starbound - Unstable\\storage\\universe\\tiles.dat").writeBytes(output)
		val reader = DataInputStream(ByteArrayInputStream(output))

		reader.skipBytes(3)

		for (y in 0 .. 31) {
			for (x in 0 .. 31) {
				println("$x $y ${reader.readShort()}")
				reader.skipBytes(29)
			}
		}

		return
	}*/

	val client = StarboundClient()

	//Starbound.addFilePath(File("./unpacked_assets/"))
	Starbound.addPakPath(File("J:\\Steam\\steamapps\\common\\Starbound\\assets\\packed.pak"))

	Starbound.initializeGame { finished, replaceStatus, status ->
		client.putDebugLog(status, replaceStatus)
	}

	client.onTermination {
		Starbound.terminateLoading = true
	}

	val ent = PlayerEntity(client.world!!)

	Starbound.onInitialize {
		var find = 0L
		var set = 0L
		var parse = 0L

		for (chunkX in 0 .. 61) {
			for (chunkY in 0 .. 61) {
				var t = System.currentTimeMillis()
				val data = db.read(byteArrayOf(1, 0, chunkX.toByte(), 0, chunkY.toByte()))
				find += System.currentTimeMillis() - t

				if (data != null) {
					val chunk = client.world!!.computeIfAbsent(ChunkPos(chunkX, chunkY))
					val inflater = Inflater()
					inflater.setInput(data)

					t = System.currentTimeMillis()
					val output = ByteArray(64_000)
					val actual = inflater.inflate(output)
					val reader = DataInputStream(ByteArrayInputStream(output))
					parse += System.currentTimeMillis() - t

					reader.skipBytes(3)

					var hitTile = false

					t = System.currentTimeMillis()

					for (y in 0 .. 31) {
						for (x in 0 .. 31) {
							val materialID = reader.readShort()
							val getMat = Starbound.tilesAccessID[materialID.toInt()]

							if (getMat != null) {
								chunk.foreground[x, y] = getMat
								hitTile = true
							}

							reader.skipBytes(5)

							val materialID2 = reader.readShort()
							val getMat2 = Starbound.tilesAccessID[materialID2.toInt()]

							if (getMat2 != null) {
								chunk.background[x, y] = getMat2
								hitTile = true
							}

							reader.skipBytes(22)
						}
					}

					set += System.currentTimeMillis() - t

					if (hitTile) {
						//println(chunk.chunk.posVector2d)
						// ent.position = chunk.chunk.posVector2d + Vector2d(16.0, 34.0)
					}
				}
			}
		}

		println("$find $set $parse")
	}

	//ent.position += Vector2d(y = 14.0, x = -10.0)
	ent.position = Vector2d(128.0 + 16.0, 672.0 + 48.0)

	client.onDrawGUI {
		client.gl.font.render("${ent.position}", y = 100f, scale = 0.25f)
		client.gl.font.render("${ent.movement.velocity}", y = 120f, scale = 0.25f)
	}

	client.onPreDrawWorld {
		//client.camera.pos.x = ent.pos.x.toFloat()
		//client.camera.pos.y = ent.pos.y.toFloat()
	}

	client.camera.pos.y = 10f

	client.gl.box2dRenderer.drawShapes = false
	client.gl.box2dRenderer.drawPairs = false
	client.gl.box2dRenderer.drawAABB = false
	client.gl.box2dRenderer.drawJoints = false

	ent.spawn()

	while (client.renderFrame()) {
		Starbound.pollCallbacks()

		//ent.think(client.frameRenderTime)
		client.camera.pos.x = ent.position.x.toFloat()
		client.camera.pos.y = ent.position.y.toFloat()

		//println(client.camera.velocity.toDoubleVector() * client.frameRenderTime * 0.1)

		//if (ent.onGround)
			//ent.velocity += client.camera.velocity.toDoubleVector() * client.frameRenderTime * 0.1

		if (client.input.KEY_LEFT_DOWN) {
			ent.movement.moveDirection = Move.MOVE_LEFT
		} else if (client.input.KEY_RIGHT_DOWN) {
			ent.movement.moveDirection = Move.MOVE_RIGHT
		} else {
			ent.movement.moveDirection = Move.STAND_STILL
		}

		if (client.input.KEY_SPACE_PRESSED && ent.movement.onGround) {
			ent.movement.requestJump()
		} else if (client.input.KEY_SPACE_RELEASED) {
			ent.movement.recallJump()
		}

		if (client.input.KEY_ESCAPE_PRESSED) {
			glfwSetWindowShouldClose(client.window, true)
		}

		ent.wantsToDuck = client.input.KEY_DOWN_DOWN

		//if (chunkA != null && glfwGetTime() < 10.0) {
		//	val tile = Starbound.getTileDefinition("alienrock")
			//chunkA!!.foreground[rand.nextInt(0, CHUNK_SIZE_FF), rand.nextInt(0, CHUNK_SIZE_FF)] = tile
		//}
	}
}
